from unittest.mock import patch

import pytest
from pytest_django.asserts import assertInHTML

from django_sample.models import Author, Book, Tag

from .example_schema import QueryExecutor

base_query = QueryExecutor().build_query(
    """
    query ChangelogQuery($page_num :Int!) {
        changelog(
            page_num:$page_num,
        ){
            has_next_page
            changelog_entries {
                model_name
                version {
                    instance
                    edited_by
                }
                eternal
        
                live_name
                edit_date
                diffs {
                    field
                    field_name
                    action
                    diffed_before
                    diffed_after
                    diffed_combined
                }
            }
        }
    }
"""
)


def test_changelog():
    data = base_query(page_num=1)


def test_entire_graphql_changelog():
    author1 = Author.objects.create(first_name="john", last_name="smith")
    author2 = Author.objects.create(first_name="jane", last_name="smith")
    book1 = Book.objects.create(author=author1, title="john's diary")

    book1_v1 = book1.versions.last()

    # refresh a new copy of this record to avoid history-row re-use
    book1.reset_version_attrs()
    book1.title = "jane's diary"
    book1.author = author2
    book1.save()

    book1_v2 = book1.versions.last()

    with patch("zeus.changelog.graphql.util.get_changelog_page_size", lambda _: 2):
        data = base_query(page_num=1)

    edit_entries = data["changelog"]["changelog_entries"]
    assert len(edit_entries) == 2

    assert edit_entries[0]["version"]["instance"] == book1_v2
    assert len(edit_entries[0]["diffs"]) == 2
    diffs_by_fields = {diff["field_name"]: diff for diff in edit_entries[0]["diffs"]}

    # field-diffs might come in any order
    assert set(diffs_by_fields.keys()) == {
        "title",
        "author",
    }
    # they have field objects as well
    assert diffs_by_fields["title"]["field"] == Book._meta.get_field("title")

    assert edit_entries[1]["version"]["instance"] == book1_v1

    # this field returns a tm() result
    assert edit_entries[1]["diffs"][0]["action"] == "Created"

    assert edit_entries[0]["eternal"] == edit_entries[1]["eternal"] == book1
    assert edit_entries[0]["model_name"] == "book"

    assert data["changelog"]["has_next_page"] == True


def test_m2m_entries():
    t1 = Tag.objects.create(name="Tag1")
    t2 = Tag.objects.create(name="Tag2")
    t3 = Tag.objects.create(name="Tag3")

    book = Book.objects.create(
        author=Author.objects.create(first_name="john", last_name="smith"),
        title="john's diary",
    )
    book.tags.add(t1, t2)

    book_v1 = book.versions.last()
    book.reset_version_attrs()

    book.tags.add(t3)
    book.tags.remove(t1)
    book.save()

    book_v2 = book.versions.last()
    data = base_query(page_num=1)
    edit_entries = data["changelog"]["changelog_entries"]
    assert len(edit_entries) == 3

    assert edit_entries[0]["version"]["instance"] == book_v2
    assert len(edit_entries[0]["diffs"]) == 1

    m2m_diff = edit_entries[0]["diffs"][0]
    assert m2m_diff["action"] == "Edited"
    assert m2m_diff["field_name"] == "tags"
    # non-added/removed tag show up without class in both before/after
    assertInHTML("<p class=''>Tag2</p>", m2m_diff["diffed_before"])
    assertInHTML("<p class=''>Tag2</p>", m2m_diff["diffed_after"])

    assertInHTML("<p class='diff_sub'>Tag1</p>", m2m_diff["diffed_before"])
    assertInHTML("<p class='diff_add'>Tag3</p>", m2m_diff["diffed_after"])

    assertInHTML("<p class=''>Tag2</p>", m2m_diff["diffed_combined"])
    assertInHTML("<p class='diff_sub'>Tag1</p>", m2m_diff["diffed_combined"])
    assertInHTML("<p class='diff_add'>Tag3</p>", m2m_diff["diffed_combined"])



def get_excluded_fields_for_model(live_model):
    if not hasattr(live_model, "excluded_diff_fields"):
        return ["id"]

    return ["id", *live_model.excluded_diff_fields]


def get_diffable_fields_for_model(live_model):
    all_fields = [*live_model._meta.fields, *live_model._meta.many_to_many]

    excluded = get_excluded_fields_for_model(live_model)

    ret = []
    for f in all_fields:
        if f.name in excluded:
            continue
        ret.append(f)

    return ret



from .changelog_entry import ChangelogEntry
from .changelog_entry_field_entry import ChangelogEntryFieldEntry
from .changelog_page import ChangelogPage
from .diff import Diff
from .version_comparison_entry import VersionComparisonEntry



import os

from zeus.i18n.text_maker import TextMakerCreator

path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
yaml_file = os.path.join(dir_path, "changelog_text.text.yaml")

tm = TextMakerCreator({"en": {}, "fr": {}}, [yaml_file]).get_tm_func()



from inspect import isgenerator

from graphql.execution.middleware import make_it_promise
from promise import is_thenable

from .utils import (
    NonSerializable,
    _stringify_internal_python_value,
    promise_from_generator,
)


def promised_generator_middleware(next, root, info, **kwargs):
    next_val = next(root, info, **kwargs)

    if not is_thenable(next_val):
        return next_val

    def handler(resolved_next_val):
        if isgenerator(resolved_next_val):
            prom = promise_from_generator(resolved_next_val)
            return prom
        return resolved_next_val

    return next_val.then(handler)


def conditonal_serialization_middleware(next, root, info, **kwargs):
    next_val = next(root, info, **kwargs)

    is_internal_field = hasattr(info.return_type, "graphene_type") and issubclass(
        info.return_type.graphene_type, NonSerializable
    )
    if not (info.context.requires_serializable and is_internal_field):
        return next_val

    if is_thenable(next_val):
        return next_val.then(lambda val: _stringify_internal_python_value(val))
    else:
        return _stringify_internal_python_value(next_val)


def get_middleware():
    return [
        promised_generator_middleware,
        conditonal_serialization_middleware,
    ]



from django.db import models
from django.db.models import F, Manager, OuterRef, QuerySet, Subquery
from django.db.models.base import ModelBase
from django.db.models.signals import m2m_changed, post_save
from django.utils import timezone

# other imports, to remove


class VersioningException(Exception):
    pass


class VersioningConfigException(VersioningException):
    pass


def find_m2m_field(from_class, to_class):
    return next(
        rel
        for rel in from_class._meta.get_fields()
        if isinstance(rel, models.ManyToManyField) and rel.related_model is to_class
    )


def on_m2m_change(instance, action, model, pk_set, **_kwargs):
    field = find_m2m_field(instance.__class__, model)

    hidden_attr = f"_{field.name}_m2m_ids"
    if not hasattr(instance, hidden_attr):

        # m2m changes, when performed on their own, wont trigger a new version. So we call .save() to force a new version
        # BUT, a form is expected to create a single version
        # so we keep track of this state manually via an attribute
        # TODO: find a way to autmatically clear this attribute
        if not (
            hasattr(instance, "_apply_changes_to_last_ver")
            and instance._apply_changes_to_last_ver
        ):
            instance.save()

        initial_ids = [obj.pk for obj in getattr(instance, field.name).all()]
        setattr(instance, hidden_attr, initial_ids)

    if action in ("post_add", "post_remove"):
        old_ids = getattr(instance, hidden_attr)
        if action == "post_remove":
            new_ids = set(old_ids) - set(pk_set)

        if action == "post_add":
            new_ids = set(old_ids).union(set(pk_set))

        new_ids = sorted(new_ids)
        setattr(instance, hidden_attr, new_ids)

        related_versions = instance.versions
        related_versions.last().set_m2m(field, new_ids)


def save_copy_post_save(sender, instance, **_kwargs):
    if hasattr(sender, "_history_class"):
        if (
            hasattr(instance, "_apply_changes_to_last_ver")
            and instance._apply_changes_to_last_ver
        ):
            sender._history_class.update_instance_version(instance)
        else:
            sender._history_class.create_from_original(instance)

        setattr(instance, "_apply_changes_to_last_ver", True)

    setattr(instance, "_apply_changes_to_last_ver", True)


class HistoryQueryset(QuerySet):
    def with_previous_version_id(self):
        # we need a non-filtered QS clone of the same type
        another_qs = HistoryQueryset(self.model, self._db)
        prev_version_id_subquery = Subquery(
            another_qs.filter(
                eternal_id=OuterRef("eternal_id"),
                business_date__lt=OuterRef("business_date"),
            )
            .order_by("-business_date")
            .values("id")[:1]
        )
        return self.annotate(previous_version_id=prev_version_id_subquery)

    def with_most_recent_version_id(self):
        most_recent_version_id_subquery = Subquery(
            self.model.objects.filter(eternal_id=OuterRef("eternal_id"))
            .order_by("-business_date")
            .values("pk")[:1]
        )
        return self.annotate(most_recent_version_id=most_recent_version_id_subquery)

    def only_most_recent_versions(self):
        return self.with_most_recent_version_id().filter(id=F("most_recent_version_id"))


class HistoryManager(Manager):
    def get_queryset(self):
        return HistoryQueryset(self.model, using=self._db)

    def latest_versions_for_eternal_ids(self, eternal_ids):

        return self.model.objects.filter(
            eternal_id__in=eternal_ids
        ).only_most_recent_versions()


def create_version_field_from_live_field(field):
    name = field.attname

    ## forces uses of id as the primary key
    if name == "id":
        new_field = models.ForeignKey(
            field.model, on_delete=models.CASCADE, related_name="versions",
        )
        return "eternal", new_field

    if isinstance(field, models.ForeignKey):
        FieldClass = field.__class__
        related_model = field.remote_field.model
        related_name = "+"
        if field.many_to_one:
            new_field = FieldClass(
                related_model,
                verbose_name=field.verbose_name,
                null=True,
                on_delete=models.SET_NULL,
                default=field.default,
                related_name=related_name,
            )
            name = field.name
        elif field.one_to_one:
            # 1-1 with live table becomes 1-many with version table
            new_field = models.ForeignKey(
                field.target_field.model,
                verbose_name=field.verbose_name,
                null=True,
                on_delete=models.SET_NULL,
                default=field.default,
                related_name=related_name,
            )
            name = field.name
        else:
            raise VersioningConfigException("Unexpected foreignkey type")

    else:
        new_field = clone_with_unique_false(field)
        new_field.primary_key = False

    new_field.choices = field.choices

    return name, new_field


def clone_with_unique_false(field):
    name, path, args, kwargs = field.deconstruct()
    if kwargs.get("unique", False):
        new_kwargs = {
            **kwargs,
            "unique": False,
        }
    else:
        new_kwargs = kwargs
    return field.__class__(*args, **new_kwargs)


class VersionModelMeta(ModelBase):
    def __new__(cls, cls_name, bases, cls_attrs, **kwargs):

        options = cls_attrs.get("Meta", None)

        if getattr(options, "abstract", False):
            return super().__new__(cls, cls_name, bases, cls_attrs, **kwargs)
        else:
            return cls._create_version_class(cls, cls_name, bases, cls_attrs, **kwargs)

    @staticmethod
    def _create_version_class(cls, cls_name, bases, cls_attrs, **kwargs):

        try:
            live_model = cls_attrs["live_model"]
        except (AttributeError, KeyError) as e:
            raise VersioningConfigException(
                "must define live_model attribute on version class"
            )

        if hasattr(live_model, "_history_class"):
            raise VersioningConfigException(
                "cannot define 2 history classes for a single model"
            )

        version_cls = super().__new__(cls, cls_name, bases, cls_attrs, **kwargs)

        versioned_fields = cls._get_versioned_fields(live_model, version_cls)
        for name, field_obj in versioned_fields.items():
            field_obj.contribute_to_class(version_cls, name)

        # many-to-many
        versioned_m2m_fields = cls._create_m2m_fields(version_cls, live_model)
        version_cls.m2m_fields = versioned_m2m_fields.values()
        for name, field_obj in versioned_m2m_fields.items():
            field_obj.contribute_to_class(version_cls, name)

        live_model._history_class = version_cls
        cls._attach_signals(live_model)
        live_model.reset_version_attrs = VersionModelMeta.reset_version_attrs

        return version_cls

    @staticmethod
    def _get_versioned_fields(live_model, version_cls):
        tracked_fields = version_cls.get_fields_to_version()

        versioned_fields = {}
        for field in tracked_fields:
            new_name, new_field = create_version_field_from_live_field(field)
            versioned_fields[new_name] = new_field

        return versioned_fields

    @staticmethod
    def _create_m2m_fields(version_cls, live_model):
        m2m_fields_to_add = {}
        for field in version_cls.get_m2m_fields_to_version():
            new_field = models.JSONField()
            m2m_fields_to_add[field.name] = new_field

        return m2m_fields_to_add

    @staticmethod
    def _attach_signals(live_model):
        post_save.connect(save_copy_post_save, live_model)

        for field in live_model._meta.many_to_many:
            through_model = getattr(live_model, field.name).through
            m2m_changed.connect(on_m2m_change, sender=through_model)

    # 'self' here refers to live model, we attach this dynamically
    def reset_version_attrs(self):
        if hasattr(self, "_apply_changes_to_last_ver"):
            del self._apply_changes_to_last_ver

        for field in self._meta.many_to_many:
            hidden_attr = f"_{field.name}_m2m_ids"
            if hasattr(self, hidden_attr):
                delattr(self, hidden_attr)


class VersionModel(models.Model, metaclass=VersionModelMeta):
    # you can subclass this class to add versioning columns (e.g. version-name, timestamp)
    class Meta:
        abstract = True

    objects = HistoryManager()

    system_date = models.DateTimeField(default=timezone.now)

    @classmethod
    def get_fields_to_version(cls):
        # override to include/exclude individual fields from the live model
        return cls.live_model._meta.fields

    @classmethod
    def get_m2m_fields_to_version(cls):
        # override to include/exclude individual fields from the live model
        return cls.live_model._meta.many_to_many

    @classmethod
    def build_from_original(cls, live_instance, m2m_dict=None):
        instance_dict = {
            f.attname: live_instance.serializable_value(f.name)
            for f in cls.get_fields_to_version()
            if not f.name in ["id"]
        }

        instance_dict.update(
            {
                f.attname: cls.serialize_m2m_ids(
                    [related.id for related in getattr(live_instance, f.name).all()]
                )
                for f in cls.m2m_fields
            }
        )

        instance_dict["eternal"] = live_instance
        ver = cls(**instance_dict)
        return ver

    @classmethod
    def create_from_original(cls, live_instance):
        ver = cls.build_from_original(live_instance)
        ver.save()
        return ver

    @classmethod
    def update_instance_version(cls, instance):
        version = instance.versions.last()
        # update all non-m2m fields
        for f in cls.live_model._meta.fields:
            if not f.name in ["id", "system_date"]:
                setattr(version, f.attname, instance.serializable_value(f.name))

        version.save()

    @staticmethod
    def serialize_m2m_ids(pk_list):
        return sorted(pk_list)

    def get_m2m_ids(self, key):
        getattr(self, key)

    def set_m2m(self, field, pk_set):
        setattr(self, field.name, self.serialize_m2m_ids(pk_set))
        self.save()

    def recreate_original(self):
        live_fields = [f for f in self.live_model._meta.fields if f.name != "id"]
        record_attrs = {f.attname: getattr(self, f.attname) for f in live_fields}

        obj = self.live_model(**record_attrs)
        obj.id = obj.pk = self.eternal_id

        # defensively remove this temporary object's save method
        def new_save():
            raise VersioningException("can't save live-obj recreated from a version")

        setattr(obj, "save", new_save)

        return obj



from django.conf import settings

import yaml


class TextEntryManager:
    def __init__(self, paths):
        self.paths = paths
        self.base_dir = settings.BASE_DIR
        self._text_entries = None

    def load(self):
        self._text_entries = {}
        for filename in self.paths:
            f = open(filename, "r")
            yaml_dict = yaml.load(f, Loader=yaml.SafeLoader)
            self._add_text_entries(yaml_dict)
            f.close()

    def _add_text_entries(self, to_add):
        for key in to_add.keys():
            if key in self._text_entries:
                raise Exception(f"text key: {key} is already defined")
        self._text_entries.update(to_add)

    def get_entry(self, key):
        return self._text_entries.get(key)

    def has_entry(self, key):
        return key in self._text_entries



from django.db.models import ForeignKey, ManyToManyField

import graphene
from graphene import DateTime, List, String

from zeus.changelog.util import get_diffable_fields_for_model
from zeus.graphql.utils import NonSerializable, non_serializable_field

from .diff import (
    Diff,
    get_display_value,
    get_field_diff_for_version_pair,
    is_field_different_accross_versions,
)


class ChangelogEntryFieldEntry(graphene.ObjectType):
    diff = graphene.Field(Diff)
    field_name = graphene.String()
    field = graphene.String()
    previous_display_value = graphene.String()
    current_display_value = graphene.String()
    has_difference = graphene.Boolean()

    @staticmethod
    def resolve_diff(parent, info):
        if parent["previous_version"] is None:
            return None

        return get_field_diff_for_version_pair(
            parent["version"],
            parent["previous_version"],
            parent["field"],
            info.context.dataloaders,
        )

    @staticmethod
    def resolve_previous_display_value(parent, info):
        if parent["previous_version"] is None:
            return None
        return get_display_value(
            parent["previous_version"], parent["field"], info.context.dataloaders
        )

    @staticmethod
    def resolve_current_display_value(parent, info):
        return get_display_value(
            parent["version"], parent["field"], info.context.dataloaders
        )

    @staticmethod
    def resolve_field(parent, info):
        return parent["field"].name

    @staticmethod
    def resolve_field_name(parent, info):
        return parent["field"].verbose_name

    @staticmethod
    def resolve_has_difference(parent, _info):
        if not parent["previous_version"]:
            return None
        return is_field_different_accross_versions(
            parent["version"], parent["previous_version"], parent["field"].name
        )



import string


def are_strings_close_enough(s1, s2):
    """
    tests that both strings are equal if you remove punctuation, space characters and ignore case
  """
    remove = string.punctuation + string.whitespace
    translate_table = str.maketrans(dict.fromkeys(remove))

    return s1.lower().translate(translate_table) == s2.lower().translate(translate_table)



import os
import re

from django.conf import settings


class TextWatcher:
    def __init__(self):
        self._observer = None
        self._entry_managers_to_watch = []
        self._is_watching = False

    def get_all_watched_files(self):
        files = set()
        for entry_manager in self._entry_managers_to_watch:
            files.update({*entry_manager.paths})
        return files

    def is_file_in_watch_scope(self, suspected_file):
        for watched_file in self.get_all_watched_files():
            if watched_file in suspected_file:
                return True

        return False

    def start_watching(self):
        if not os.environ.get("RUN_MAIN", None):
            # runserver spawns two processes, the initial one watches code and reloads files
            # the other one changes PID and has an env RUN_MAIN=True
            return

        from watchdog.events import RegexMatchingEventHandler
        from watchdog.observers import Observer

        watcher = self

        class RestartOnModifiedEventHandler(RegexMatchingEventHandler):
            def on_modified(self, event):
                # TODO check that the eveng matches a file in get_all_watched_files()

                if watcher.is_file_in_watch_scope(event.src_path):
                    print("=========== reloading yaml text ===========")
                    try:
                        for entry_manager in watcher._entry_managers_to_watch:
                            entry_manager.load()
                    except Exception as e:
                        print(e)
                        print(
                            "text will not work until the above exception is fixed"
                            "\n"
                            "==========="
                        )

        self._observer = Observer()
        self._observer.daemon = True
        event_handler = RestartOnModifiedEventHandler(
            regexes=[r".*.text.yaml"], ignore_directories=True
        )
        self._observer.schedule(
            event_handler, os.path.abspath(settings.BASE_DIR), recursive=True
        )
        self._observer.start()

    def add_entry_watcher(self, entry_manager):
        if self._observer:
            # if already watching, restart the watcher with the new entry manager
            self._stop_watching()
            self._entry_managers_to_watch.append(entry_manager)
            self.start_watching()

        else:
            self._entry_managers_to_watch.append(entry_manager)

    def _stop_watching(self):
        if self._observer:
            self._observer.stop()
            self._observer = None


text_watcher = TextWatcher()



import difflib
from itertools import groupby

ADD_HTML_CLASS = "diff_add"
RM_HTML_CLASS = "diff_sub"


class DiffAdd(str):
    @staticmethod
    def combine_to_html(strings):
        return f"<span class='{ADD_HTML_CLASS}'>{' '.join(strings)}</span>"


class DiffRemove(str):
    @staticmethod
    def combine_to_html(strings):
        return f"<span class='{RM_HTML_CLASS}'>{' '.join(strings)}</span>"


class DiffNull(str):
    # these are normal, non-diff strings
    # class is just here for interface's sake
    @staticmethod
    def combine_to_html(strings):
        return " ".join(strings)


def text_compare_inline(before, after):
    """returns a 3-tuple of text"""
    joint = []
    old = []
    new = []

    for word in list(difflib.ndiff(before.split(), after.split())):
        if word.startswith("+ "):
            addition = DiffAdd(word[2:])
            new.append(addition)
            joint.append(addition)
        elif word.startswith("- "):
            removal = DiffRemove(word[2:])
            old.append(removal)
            joint.append(removal)
        elif word.startswith("?"):
            pass
        else:
            word = DiffNull(word)
            old.append(word)
            new.append(word)
            joint.append(word)

    # we want consecutive removes/adds to be in the same <span>,
    # recall that groupby groups consecutive elements together
    joint_formated = " ".join(
        cls.combine_to_html(strings)
        for (cls, strings) in groupby(joint, lambda diffstr: diffstr.__class__)
    )
    old_formated = " ".join(
        cls.combine_to_html(strings)
        for (cls, strings) in groupby(old, lambda diffstr: diffstr.__class__)
    )
    new_formated = " ".join(
        cls.combine_to_html(strings)
        for (cls, strings) in groupby(new, lambda diffstr: diffstr.__class__)
    )

    return (
        joint_formated,
        old_formated,
        new_formated,
    )


def list_diff(before_list, after_list):
    """
        expects 2 lists of strings whose intersections are sorted identically
        This only ever makes sense if lists contain unique elements
    """

    added = set(after_list) - set(before_list)
    removed = set(before_list) - set(after_list)
    combined_list = list({*before_list, *after_list})

    annotated_before_list = []
    annotated_after_list = []
    annotated_combined_list = []

    for string in combined_list:
        is_removed = is_added = False

        if string in added:
            html_class = ADD_HTML_CLASS
            is_added = True
        elif string in removed:
            html_class = RM_HTML_CLASS
            is_removed = True
        else:
            html_class = ""

        annotated_str = f"<p class='{html_class}'>{string}</p>"

        annotated_combined_list.append(annotated_str)

        if is_added:
            annotated_after_list.append(annotated_str)
        elif is_removed:
            annotated_before_list.append(annotated_str)
        else:
            # in both groups -> add it to both
            annotated_before_list.append(annotated_str)
            annotated_after_list.append(annotated_str)

    return (
        "".join(annotated_combined_list),
        "".join(annotated_before_list),
        "".join(annotated_after_list),
    )



import inspect

from django.db.models import ForeignKey, JSONField, ManyToManyField
from django.utils.functional import cached_property
from django.utils.html import escape

import graphene

from promise import Promise

from zeus.changelog.diff_utils import list_diff, text_compare_inline
from zeus.changelog.text import tm
from zeus.graphql.dataloader import PrimaryKeyDataLoaderFactory
from zeus.graphql.utils import NonSerializable, genfunc_to_prom, non_serializable_field


# this is used by the parent resolver and fed to the below graphene type
class DiffObject:
    def __init__(self, current_version, previous_version, field_obj, dataloader_cache):
        self.current_version = current_version
        self.previous_version = previous_version
        self.field = field_obj
        self.dataloader_cache = dataloader_cache


class ScalarDiffObject(DiffObject):
    @cached_property
    def _diffs(self):
        prev_db_value = self.previous_version.serializable_value(self.field.name)
        current_db_value = self.current_version.serializable_value(self.field.name)

        if self.field.choices:
            # if a field is a choice field e.g. chars or ints used to represent a list of choices,
            # then its value is just that database, non-bilingual char/int value
            # fortunately model instances provide a hook to get the displayed, translated value
            choices_by_attr_value = dict(self.field.choices)
            if not choices_by_attr_value.get(None, None):
                choices_by_attr_value[None] = tm("empty")
            previous_value = choices_by_attr_value.get(prev_db_value, prev_db_value)
            current_value = choices_by_attr_value.get(current_db_value, current_db_value)

        else:  # just use the normal attribute
            previous_value = prev_db_value
            current_value = current_db_value

        joint, before, after = text_compare_inline(
            get_str_val(previous_value), get_str_val(current_value),
        )

        return (joint, before, after)

    def get_combined_diff(self):
        return self._diffs[0]

    def get_before_diff(self):
        return self._diffs[1]

    def get_after_diff(self):
        return self._diffs[2]


class CreateDiff:
    def __init__(self):
        self.field = None

    def _get_created_str(self):
        return escape(tm("created"))

    def get_before_diff(self):
        return ""

    def get_after_diff(self):
        return self._get_created_str()

    def get_combined_diff(self):
        return self._get_created_str()


class DeleteDiff:
    def __init__(self):
        self.field = None

    def _get_deleted_str(self):
        return escape(tm("deleted"))

    def get_before_diff(self):
        return ""

    def get_after_diff(self):
        return self._get_deleted_str()

    def get_combined_diff(self):
        return self._get_deleted_str()


class AsyncDiffObject(DiffObject):
    def _compute_diffs(self):
        raise NotImplementedError()

    @genfunc_to_prom
    def _get_diffs(self):
        if hasattr(self, "_diff_result"):
            return self._diff_result

        diffs = yield self._compute_diffs()
        self._diff_result = diffs
        return diffs

    @genfunc_to_prom
    def get_combined_diff(self):
        diffs = yield self._get_diffs()
        return diffs[0]

    @genfunc_to_prom
    def get_before_diff(self):
        diffs = yield self._get_diffs()
        return diffs[1]

    @genfunc_to_prom
    def get_after_diff(self):
        diffs = yield self._get_diffs()
        return diffs[2]


class M2MDiffObject(AsyncDiffObject):
    @genfunc_to_prom
    def _compute_diffs(self):
        prev_id_list = self.previous_version.serializable_value(self.field.name)
        current_id_list = self.current_version.serializable_value(self.field.name)

        related_model = self.field.related_model
        related_dataloader_cls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(
            related_model
        )
        related_dataloader = related_dataloader_cls(self.dataloader_cache)

        prev_instances, current_instances = yield Promise.all(
            [
                related_dataloader.load_many(prev_id_list),
                related_dataloader.load_many(current_id_list),
            ]
        )

        get_name = lambda inst: inst.name
        before_list = sorted([*prev_instances], key=get_name)
        after_list = sorted([*current_instances], key=get_name)

        joint, before, after = list_diff(before_list, after_list)

        return (joint, before, after)


class ForeignKeyDiffObject(AsyncDiffObject):
    @genfunc_to_prom
    def _compute_diffs(self):
        prev_db_value = self.previous_version.serializable_value(self.field.name)
        current_db_value = self.current_version.serializable_value(self.field.name)

        related_model = self.field.remote_field.model
        related_dataloader_cls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(
            related_model
        )
        dataloader = related_dataloader_cls(self.dataloader_cache)

        if prev_db_value is None:
            previous_instance = None
            current_instance = yield dataloader.load(current_db_value)
        elif current_db_value is None:
            previous_instance = yield dataloader.load(prev_db_value)
            current_instance = None
        else:
            previous_instance, current_instance = yield dataloader.load_many(
                [prev_db_value, current_db_value]
            )

        joint, before, after = text_compare_inline(
            get_str_val(previous_instance), get_str_val(current_instance),
        )

        return joint, before, after


def is_field_different_accross_versions(current_version, previous_version, field_name):
    current_db_value = current_version.serializable_value(field_name)
    prev_db_value = previous_version.serializable_value(field_name)

    field_obj = current_version._meta.get_field(field_name)
    if isinstance(field_obj, JSONField):
        return current_db_value != prev_db_value

    if current_db_value == prev_db_value or (
        # consider "" vs. None to be non-diff worthy
        {current_db_value, prev_db_value}.issubset({None, ""})
    ):
        return False
    return True


def get_field_diff_for_version_pair(
    current_version, previous_version, field_obj, dataloaders
):
    if not is_field_different_accross_versions(
        current_version, previous_version, field_obj.name
    ):
        return None

    if isinstance(field_obj, ManyToManyField):
        DiffClass = M2MDiffObject
    elif isinstance(field_obj, ForeignKey):
        DiffClass = ForeignKeyDiffObject
    else:
        DiffClass = ScalarDiffObject

    diff_obj = DiffClass(current_version, previous_version, field_obj, dataloaders)
    return diff_obj


def get_str_val(fetched_field_value):
    if isinstance(fetched_field_value, str):
        return escape(fetched_field_value)
    elif fetched_field_value in (None, ""):
        return escape(tm("empty"))
    if isinstance(fetched_field_value, object):
        return escape(fetched_field_value.__str__())


@genfunc_to_prom
def m2m_display_value(version, field_obj, dataloader_cache):
    id_list = version.get_m2m_ids(field_obj.name)
    if not id_list:
        return tm("empty")
    related_model = field_obj.related_model
    related_dataloader_cls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(
        related_model
    )
    related_dataloader = related_dataloader_cls(dataloader_cache)
    related_instances = yield related_dataloader.load_many(id_list)

    sorted_names = sorted([inst.name for inst in related_instances])
    return "".join([f"<p>{name}</p>" for name in sorted_names])


@genfunc_to_prom
def foreign_key_display_value(version, field_obj, dataloader_cache):
    related_id = version.serializable_value(field_obj.name)
    if related_id is None:
        return tm("empty")
    related_model = field_obj.related_model
    related_dataloader_cls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(
        related_model
    )
    related_dataloader = related_dataloader_cls(dataloader_cache)

    related_instance = yield related_dataloader.load(related_id)
    return get_str_val(related_instance)


def get_display_value(version, field_obj, dataloader_cache):
    if isinstance(field_obj, ManyToManyField):
        return m2m_display_value(version, field_obj, dataloader_cache)

    if isinstance(field_obj, ForeignKey):
        return foreign_key_display_value(version, field_obj, dataloader_cache)

    if field_obj.choices:
        db_value = version.serializable_value(field_obj.name)
        choices_by_attr_value = dict(field_obj.choices)
        if not choices_by_attr_value.get(None, None):
            choices_by_attr_value[None] = tm("empty")
        str_val = choices_by_attr_value.get(db_value, db_value)
        return get_str_val(str_val)

    else:
        # use normal attribute
        val = getattr(version, field_obj.name)
        if not val:
            return tm("empty")
        return val


class Diff(graphene.ObjectType):
    diffed_before = graphene.String()
    diffed_after = graphene.String()
    diffed_combined = graphene.String()
    action = graphene.String()
    field_name = graphene.String()
    field = NonSerializable()

    @staticmethod
    def resolve_field(parent, _info):
        return parent.field

    @staticmethod
    def resolve_field_name(parent, _info):
        field_obj = parent.field
        if field_obj is None:
            return None
        return field_obj.name

    @staticmethod
    def resolve_action(parent, _info):
        if isinstance(parent, CreateDiff):
            return tm("created")
        elif isinstance(parent, DeleteDiff):
            return tm("deleted")
        else:
            return tm("edited")

    @staticmethod
    def resolve_diffed_before(parent, _info):
        return parent.get_before_diff()

    @staticmethod
    def resolve_diffed_after(parent, _info):
        return parent.get_after_diff()

    @staticmethod
    def resolve_diffed_combined(parent, _info):
        return parent.get_combined_diff()



import urllib

from django.conf import settings
from django.http.response import HttpResponseRedirect


class MustBeLoggedInMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.user.is_authenticated:
            return self.get_response(request)

        elif "/login" not in request.path.lower():
            qs_params = dict(next=request.build_absolute_uri())
            querystring = urllib.parse.urlencode(qs_params)
            return HttpResponseRedirect(f"{settings.LOGIN_URL}?{querystring}")

        else:
            return self.get_response(request)



from django.conf import settings
from django.db import connections
from django.db.models import signals


def mark_whodid(user, sender, instance, **kwargs):
    if hasattr(sender, "edited_by"):
        instance.edited_by = user


class WhodidMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.method not in ("GET", "HEAD", "OPTIONS", "TRACE"):
            if hasattr(request, "user") and request.user.is_authenticated:
                user = request.user
            else:
                user = None

            def wrapped_mark_whodid(*args, **kwargs):
                return mark_whodid(user, *args, **kwargs)

            signals.pre_save.connect(
                wrapped_mark_whodid, dispatch_uid=(self.__class__, request,), weak=False
            )

        response = self.get_response(request)

        signals.pre_save.disconnect(dispatch_uid=(self.__class__, request,))

        return response



import graphene

from zeus.graphql.types.version import Version
from zeus.graphql.utils import NonSerializable, non_serializable_field


class HasNonSerializableRecordMixin(graphene.ObjectType):
    class Meta:
        abstract = True

    record = graphene.Field(NonSerializable)
    most_recent_version = graphene.Field(Version)

    @staticmethod
    def get_orm_instance_from_resolver_parent(resolver_parent):
        return resolver_parent

    @classmethod
    @non_serializable_field
    def resolve_record(cls, parent, _info):
        return cls.get_orm_instance_from_resolver_parent(parent)



from collections import defaultdict


def group_by_iteritems(iterable, key=lambda x: x):
    d = defaultdict(list)
    for item in iterable:
        d[key(item)].append(item)
    return d.items()


def group_by(iterable, key):
    return dict(group_by_iteritems(iterable, key))



from django.contrib.auth.models import User
from django.urls import reverse

import pytest

from django_sample.models import Author, Book


def test_edit_book_has_edited_by_prop(client):
    user = User.objects.create(username="abc")
    client.force_login(user)
    author = Author.objects.create(first_name="john")
    a_book = Book.objects.create(author=author, title="name1")
    a_book.reset_version_attrs()

    url = reverse("edit-book", args=[a_book.pk])
    client.post(url, data={"title": "name2"})

    assert a_book.versions.count() == 2
    assert a_book.versions.first().edited_by is None
    assert a_book.versions.last().edited_by == user



from collections import defaultdict

import pytest
from promise import Promise

from django_sample.data_factories import AuthorFactory, BookFactory, TagFactory
from django_sample.models import Book, Tag
from zeus.django.query_counting import assert_max_queries
from zeus.graphql.dataloader import PrimaryKeyDataLoaderFactory, SingletonDataLoader
from zeus.vanilla import flatten, group_by


def test_pk_loader_factory(dataloader_test):
    b1 = BookFactory()
    b2 = BookFactory()
    dataloader_cache = {}
    BookLoaderCls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(Book)
    with assert_max_queries(0):
        book1_prom = BookLoaderCls(dataloader_cache).load(b1.pk)
        book2_prom = BookLoaderCls(dataloader_cache).load(b2.pk)

    with assert_max_queries(1):
        resolved_b1, resolved_b2 = Promise.all([book1_prom, book2_prom]).get()

        assert b1 == resolved_b1
        assert b2 == resolved_b2


class BooksForAuthorLoader(SingletonDataLoader):
    def batch_load(self, author_ids):
        books = Book.objects.filter(author_id__in=author_ids)
        by_author = group_by(books, lambda b: b.author_id)
        return [by_author.get(id, []) for id in author_ids]


class TagsForBookLoader(SingletonDataLoader):
    def batch_load(self, book_ids):
        books = Book.objects.filter(id__in=book_ids).prefetch_related("tags")
        by_id = defaultdict(list)
        for b in books:
            by_id[b.id].extend(b.tags.all())

        return [by_id.get(id) for id in book_ids]


class TagsForAuthorLoader(SingletonDataLoader):
    def batch_load(self, author_ids):
        book_groups = yield BooksForAuthorLoader(
            self.dataloader_instance_cache
        ).load_many(author_ids)
        books = flatten(book_groups)
        book_ids_by_author = group_by(books, lambda b: b.author_id)
        book_ids = [b.id for b in books]

        tag_groups = yield TagsForBookLoader(self.dataloader_instance_cache).load_many(
            book_ids
        )
        tags_by_book_id = dict(zip(book_ids, tag_groups))

        tags_by_author_id = defaultdict(set)
        for b in books:
            tags = tags_by_book_id[b.id]
            tags_by_author_id[b.author_id].update(tags)

        return [tags_by_author_id[id] for id in author_ids]


def test_composed_dataloader(dataloader_test):
    a1 = AuthorFactory()
    a2 = AuthorFactory()
    a3 = AuthorFactory()

    t1 = TagFactory()
    t2 = TagFactory()
    t3 = TagFactory()

    b1_1 = BookFactory(author=a1, tags=[t1, t3])
    b1_2 = BookFactory(author=a1)
    b2_1 = BookFactory(author=a2, tags=[t2])
    b2_2 = BookFactory(author=a2, tags=[t1])

    dataloader_cache = {}

    with assert_max_queries(0):
        a1_prom = TagsForAuthorLoader(dataloader_cache).load(a1.pk)
        a2_prom = TagsForAuthorLoader(dataloader_cache).load(a2.pk)
        a3_prom = TagsForAuthorLoader(dataloader_cache).load(a3.pk)

    with assert_max_queries(3):
        resolved_a1, resolved_a2, resolved_a3 = Promise.all(
            [a1_prom, a2_prom, a3_prom]
        ).get()

        assert resolved_a1 == {t1, t3}
        assert resolved_a2 == {t1, t2}
        assert resolved_a3 == set()



import importlib

from django.db import connection, models, transaction
from django.db.models.base import ModelBase
from django.forms import ModelForm

import pytest

from zeus.versioning.core import VersionModel


@pytest.fixture(scope="module")
def register_model(django_db_setup, django_db_blocker):
    registered_models = []

    def register(model_cls):
        registered_models.append(model_cls)
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(model_cls)

    with django_db_blocker.unblock():
        try:
            with transaction.atomic():
                yield register

                with connection.schema_editor() as schema_editor:
                    for model in registered_models:
                        schema_editor.delete_model(model)

                raise Exception(
                    "any exception will cause the transaction to be rolled back"
                )

        # pylint: disable="broad-except"
        except Exception:
            pass


# we use module-scope because django complains if you register the same model twice
@pytest.fixture(scope="module")
def common(register_model):

    # This can be any module in <installed_app>
    # when extracting to own package
    # tests will against on an 'example project' with an empty app
    module = "django_sample.models"

    class MyGroupLookup(models.Model):
        __module__ = module
        name = models.CharField(max_length=20)

    class MyLiveModel(models.Model):
        __module__ = module
        name = models.CharField(max_length=20)
        groups = models.ManyToManyField(MyGroupLookup)
        favorite_group = models.ForeignKey(
            MyGroupLookup, null=False, on_delete=models.CASCADE
        )

    class MyModelVersion(VersionModel):
        live_model = MyLiveModel
        __module__ = module

    register_model(MyGroupLookup)
    register_model(MyLiveModel)
    register_model(MyModelVersion)

    group_1 = MyGroupLookup.objects.create(name="group1")
    group_2 = MyGroupLookup.objects.create(name="group2")
    group_3 = MyGroupLookup.objects.create(name="group3")

    class NameSpace:
        group1 = group_1
        group2 = group_2
        group3 = group_3
        GroupLookup = MyGroupLookup
        LiveModel = MyLiveModel
        VersionModel = MyModelVersion

    return NameSpace


def test_create_dynamic_model(common):

    live_inst_1 = common.LiveModel.objects.create(name="v1", favorite_group=common.group1)
    assert live_inst_1.versions.count() == 1
    v1 = live_inst_1.versions.last()
    assert v1.eternal == live_inst_1
    assert v1.name == "v1"
    assert v1.favorite_group_id == common.group1.id
    assert v1.groups == []

    live_inst_1.reset_version_attrs()
    live_inst_1.groups.add(common.group1)
    live_inst_1.save()
    assert live_inst_1.versions.count() == 2
    v2 = live_inst_1.versions.last()
    assert v2.groups == [common.group1.pk]

    v1.refresh_from_db()
    assert v1.groups == []


def test_m2m_add(common):
    obj = common.LiveModel.objects.create(name="v1", favorite_group=common.group1)
    obj.reset_version_attrs()

    obj.groups.add(common.group1)

    # check og version wasn't modified
    assert obj.versions.count() == 2
    assert obj.versions.first().groups == []
    assert obj.versions.last().groups == [common.group1.pk]


def test_m2m_rm(common):
    obj = common.LiveModel.objects.create(name="v1", favorite_group=common.group1)
    obj.groups.add(common.group1)

    assert obj.versions.count() == 1
    assert obj.versions.last().groups == [common.group1.pk]

    obj.reset_version_attrs()
    obj.groups.remove(common.group1)

    # check og version wasn't modified
    assert obj.versions.count() == 2
    assert obj.versions.first().groups == [common.group1.pk]
    assert obj.versions.last().groups == []


def test_m2m_add_and_rm(common):
    obj = common.LiveModel.objects.create(name="v1", favorite_group=common.group1)
    obj.groups.add(common.group1)

    assert obj.versions.count() == 1
    assert obj.versions.last().groups == [common.group1.pk]

    obj.reset_version_attrs()
    obj.groups.remove(common.group1)
    obj.groups.add(common.group2)

    assert obj.versions.count() == 2
    assert obj.versions.first().groups == [common.group1.pk]
    assert obj.versions.last().groups == [common.group2.pk]


def test_create_then_m2m_edit_doesnt_add_version(common):
    obj = common.LiveModel.objects.create(name="v1", favorite_group=common.group1)
    obj.groups.add(common.group1)
    assert obj.versions.count() == 1
    assert obj.versions.last().groups == [common.group1.pk]


def test_scalar_edit_then_m2m_edit_only_adds_one_version(common):
    obj = common.LiveModel.objects.create(name="name1", favorite_group=common.group1)
    obj.reset_version_attrs()

    obj.name = "name2"
    obj.save()
    obj.groups.add(common.group1)

    assert obj.versions.count() == 2
    assert obj.versions.first().name == "name1"
    assert obj.versions.first().groups == []
    assert obj.versions.last().name == "name2"
    assert obj.versions.last().groups == [common.group1.pk]


def test_m2m_change_then_scalar_change_only_adds_one_version(common):
    obj = common.LiveModel.objects.create(name="name1", favorite_group=common.group1)
    obj.reset_version_attrs()

    obj.groups.add(common.group1)
    obj.name = "name2"
    obj.save()

    assert obj.versions.count() == 2
    assert obj.versions.first().name == "name1"
    assert obj.versions.first().groups == []
    assert obj.versions.last().name == "name2"
    assert obj.versions.last().groups == [common.group1.pk]


def test_create_via_model_form(common):
    class Form(ModelForm):
        class Meta:
            model = common.LiveModel
            fields = [
                "name",
                "favorite_group",
                "groups",
            ]

    form_data = {
        "name": "v1",
        "favorite_group": common.group1.pk,
        "groups": [common.group1.pk],
    }

    obj = Form(form_data).save()
    assert obj.pk
    assert obj.versions.count() == 1
    v1 = obj.versions.last()
    assert v1.name == "v1"
    assert v1.groups == [common.group1.pk]



from django.db.models import ForeignKey, ManyToManyField

import graphene
from graphene import DateTime, List, String

from zeus.changelog.graphql.types.version import Version
from zeus.changelog.util import get_diffable_fields_for_model
from zeus.graphql.dataloader import PrimaryKeyDataLoaderFactory
from zeus.graphql.utils import NonSerializable, non_serializable_field

from .diff import CreateDiff, DeleteDiff, Diff, get_field_diff_for_version_pair


class VersionComparisonEntry(graphene.ObjectType):
    model_name = graphene.String()

    @staticmethod
    def resolve_model_name(parent, info):
        return parent["model"].live_model.__name__

    eternal = NonSerializable()

    @staticmethod
    def resolve_eternal(parent, info):
        model = parent["model"].live_model
        id = parent["eternal_id"]
        LoaderCls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(model)
        return LoaderCls(info.context.dataloaders).load(id)

    left_version = graphene.Field(Version)

    @staticmethod
    def resolve_left_version(parent, info):
        model = parent["model"]
        id = parent["left_id"]

        if id is None:
            return None

        LoaderCls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(model)
        version = yield LoaderCls(info.context.dataloaders).load(id)
        return {"instance": version}

    right_version = graphene.Field(Version)

    @staticmethod
    def resolve_right_version(parent, info):
        model = parent["model"]
        id = parent["right_id"]

        if id is None:
            return None

        LoaderCls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(model)
        version = yield LoaderCls(info.context.dataloaders).load(id)
        return {"instance": version}

    diffs = List(Diff)

    @staticmethod
    def resolve_diffs(parent, info):
        if not parent["right_id"]:
            return [DeleteDiff()]
        if not parent["left_id"]:
            return [CreateDiff()]

        history_model = parent["model"]
        live_model = history_model.live_model
        fields_to_diff = get_diffable_fields_for_model(live_model)

        LoaderCls = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(history_model)
        loader_inst = LoaderCls(info.context.dataloaders)
        left_ver = yield loader_inst.load(parent["left_id"])
        right_ver = yield loader_inst.load(parent["right_id"])

        field_diff_objs = []
        for f in fields_to_diff:
            field_diff_obj = get_field_diff_for_version_pair(
                right_ver, left_ver, f, info.context.dataloaders
            )
            if field_diff_obj is not None:
                field_diff_objs.append(field_diff_obj)

        return field_diff_objs



from collections import defaultdict

import graphene
from graphene import Boolean, DateTime, Int, List, String

from zeus.changelog.consecutive_versions_fetcher import ConsecutiveVersionsFetcher
from zeus.changelog.graphql.types import ChangelogPage
from zeus.changelog.util import get_diffable_fields_for_model


def create_model_enum_type(models, enum_name="ChangelogModels"):
    model_names = [m.__name__ for m in models]
    ModelNameEnum = graphene.Enum(enum_name, [(v, v) for v in model_names])
    return ModelNameEnum


def create_model_field_enum_type(models, enum_name="ChangelogModelFields"):
    model_field_enum_vals = []

    for model in models:
        for field in get_diffable_fields_for_model(model):
            if field.name == "id":
                continue

            model_field_enum_vals.append(f"{model.__name__}__{field.name}")

    ModelFieldsEnum = graphene.Enum(enum_name, [(v, v) for v in model_field_enum_vals])
    return ModelFieldsEnum


def convert_enum_list_to_dict(models, enum_list):
    if not enum_list or not models:
        return None
    models_by_name = {model.__name__: model for model in models}
    parsed_pairs = [enum_val.split("__") for enum_val in enum_list]
    fields_by_model = defaultdict(list)
    for (model_name, field_name) in parsed_pairs:
        model_cls = models_by_name[model_name]
        fields_by_model[model_cls].append(field_name)

    return fields_by_model


# as a function purely for mocking purposes
def get_changelog_page_size(page_size):
    return page_size


def create_standard_changelog_graphql_mixin(
    diffable_models,
    models_enum=None,
    model_fields_enum=None,
    page_size=50,
    field_name="changelog",
):
    # will use models_enum and model_fields_enum if provided, otherwise generates those

    if models_enum is None:
        model_enum = create_model_enum_type(diffable_models)
    if model_fields_enum is None:
        model_fields_enum = create_model_field_enum_type(diffable_models)

    field = graphene.Field(
        ChangelogPage,
        page_num=Int(required=True),
        models=List(model_enum, required=False),
        user_ids=List(Int, required=False),
        fields=List(model_fields_enum, required=False),
        exclude_create=Boolean(required=False),
        only_creates=Boolean(required=False),
        start_date=DateTime(required=False),
        end_date=DateTime(required=False),
    )

    def resolver(
        _parent,
        _info,
        page_num=None,
        models=None,
        user_ids=None,
        fields=None,
        exclude_create=False,
        only_creates=False,
        start_date=None,
        end_date=None,
    ):
        if models and fields:
            raise Exception(
                "use models or fields but not both - TODO: split this up in 2 different fields"
            )

        if models:
            model_classes = [m for m in diffable_models if m.__name__ in models]
        else:
            model_classes = [*diffable_models]

        if fields:
            fields_by_model = convert_enum_list_to_dict(model_classes, fields)
        else:
            fields_by_model = None

        fetcher = ConsecutiveVersionsFetcher(
            page_size=get_changelog_page_size(page_size),
            page_num=page_num,
            models=model_classes,
            user_ids=user_ids,
            exclude_create=exclude_create,
            only_creates=only_creates,
            fields_by_model=fields_by_model,
            start_date=start_date,
            end_date=end_date,
        )

        return {"fetcher": fetcher}

    return type(
        "StandardChangelogMixin",
        (graphene.ObjectType,),
        {field_name: field, f"resolve_{field_name}": staticmethod(resolver)},
    )



from inspect import isgeneratorfunction

from promise import Promise, is_thenable
from promise.dataloader import DataLoader

from .utils import genfunc_to_prom, promise_from_generator


class BaseDataloader(DataLoader):
    # shortcut for instance methods to turn generator instances into promises
    def batch_load_fn(self, keys):
        func = self.batch_load
        if isgeneratorfunction(func):
            func = genfunc_to_prom(func)

        result = func(keys)

        if not is_thenable(result):
            # batch_load_fn must always return a promise
            return Promise.resolve(result)

        return result


class SingletonDataLoader(BaseDataloader):
    dataloader_instance_cache = None

    def __new__(cls, dataloader_instance_cache):
        if cls not in dataloader_instance_cache:
            dataloader_instance_cache[cls] = super().__new__(cls)
        loader = dataloader_instance_cache[cls]
        assert isinstance(loader, cls)
        return loader

    def __init__(self, dataloader_instance_cache):
        if self.dataloader_instance_cache != dataloader_instance_cache:
            self.dataloader_instance_cache = dataloader_instance_cache
            super().__init__()


class AbstractModelByIdLoader(SingletonDataLoader):
    model = None  # override this part

    @classmethod
    def batch_load(cls, ids):
        records = list(cls.model.objects.filter(id__in=ids))
        by_id = {record.id: record for record in records}
        return [by_id[id] for id in ids]


class PrimaryKeyDataLoaderFactory:
    """
        This ensures the same _class_ for a single model can only be created once. 
        This is because some consumers dynamically create dataloaders based on models not yet known
    """

    dataloader_classes_by_model = {}

    @staticmethod
    def _create_dataloader_cls_for_model(model_cls):
        return type(
            f"{model_cls.__name__}ByIDLoader",
            (AbstractModelByIdLoader,),
            dict(model=model_cls),
        )

    @classmethod
    def get_model_by_id_loader(cls, model_cls):
        if model_cls in cls.dataloader_classes_by_model:
            return cls.dataloader_classes_by_model[model_cls]
        else:
            loader = cls._create_dataloader_cls_for_model(model_cls)
            cls.dataloader_classes_by_model[model_cls] = loader
            return loader



from django.contrib.auth import get_user_model

import graphene

from zeus.graphql.dataloader import PrimaryKeyDataLoaderFactory
from zeus.graphql.utils import NonSerializable, non_serializable_field


class Version(graphene.ObjectType):
    edited_by = NonSerializable()
    instance = NonSerializable()

    @staticmethod
    def resolve_edited_by(parent, info):
        UserModel = get_user_model()
        UserDataloader = PrimaryKeyDataLoaderFactory.get_model_by_id_loader(UserModel)
        user_dataloader = UserDataloader(info.context.dataloaders)
        return user_dataloader.load(parent["instance"].edited_by_id)

    @staticmethod
    def resolve_instance(parent, _info):
        return parent["instance"]



import operator
from collections import defaultdict
from functools import reduce

from django.core.paginator import Paginator
from django.db.models import CharField, F, OuterRef, Q, Subquery, Value
from django.utils.functional import cached_property

from zeus.django.model_utils import anyQ, neverQ

from .util import get_diffable_fields_for_model


# this class is used by the parent resolver. should we move it elsewhere ?
class ConsecutiveVersionsFetcher:
    def __init__(
        self,
        page_size,
        page_num,
        models,
        user_ids=None,
        exclude_create=False,
        only_creates=False,
        fields_by_model=None,
        start_date=None,
        end_date=None,
    ):
        if only_creates and (fields_by_model or exclude_create):
            raise Exception(
                "cant only show creations while excluding them or comparing specific fields"
            )

        self.page_num = page_num
        self.models = models
        self.page_size = page_size

        self.user_ids = user_ids
        self.fields_by_model = fields_by_model
        self.exclude_create = exclude_create or fields_by_model is not None
        self.only_creates = only_creates
        self.start_date = start_date
        self.end_date = end_date

    @staticmethod
    def unionize_querysets(qs1, qs2):
        return qs1.union(qs2)

    def get_base_version_qs_for_single_model(self, live_model):
        """
            You can override this to allow for custom extra filtering
        """
        history_model = live_model._history_class
        return history_model.objects.all()

    def _get_values_qs_for_single_model(self, live_model):
        history_model = live_model._history_class

        base_qs = self.get_base_version_qs_for_single_model(live_model)

        user_filter = anyQ
        if self.user_ids:
            user_filter = Q(edited_by_id__in=self.user_ids)
        qs = base_qs.with_previous_version_id().filter(user_filter)

        if self.exclude_create:
            qs = qs.filter(previous_version_id__isnull=False)

        if self.only_creates:
            qs = qs.filter(previous_version_id__isnull=True)

        if self.start_date:
            qs = qs.filter(business_date__gte=self.start_date)

        if self.end_date:
            qs = qs.filter(business_date__lte=self.end_date)

        if self.fields_by_model and self.fields_by_model.get(live_model, None):
            field_names = self.fields_by_model[live_model]
            # also filter out creations when comparing specific fields
            qs = qs.filter(previous_version_id__isnull=False)

            field_objs = [
                f
                for f in get_diffable_fields_for_model(live_model)
                if f.name in field_names
            ]

            get_annotation_name = lambda field: f"_previous_{field.name}"
            for f in field_objs:
                prev_field_value_subquery = Subquery(
                    history_model.objects.filter(
                        id=OuterRef("previous_version_id")
                    ).values(f.attname)[:1]
                )

                qs = qs.annotate(**{get_annotation_name(f): prev_field_value_subquery})

            # once annotated, we can filter on any OR differences
            # the difference doesn't seem to check for nulls vs. values, so we check that manually
            field_difference_filters = [
                (
                    ~Q(**{get_annotation_name(field): F(field.attname)})
                    | (
                        Q(**{f"{get_annotation_name(field)}__isnull": False})
                        & Q(**{f"{field.attname}__isnull": True})
                    )
                    | (
                        Q(**{f"{get_annotation_name(field)}__isnull": True})
                        & Q(**{f"{field.attname}__isnull": False})
                    )
                )
                for field in field_objs
            ]
            combined_filter = reduce(operator.__or__, field_difference_filters, neverQ)

            qs = qs.filter(combined_filter)

        qs = (
            qs.annotate(model_name=Value(live_model.__name__, CharField()))
            .only(
                "business_date",
                "system_date",
                "id",
                "eternal_id",
                "model_name",
                "previous_version_id",
            )
            .order_by("-business_date")
            .values(
                "business_date",
                "system_date",
                "id",
                "eternal_id",
                "model_name",
                "previous_version_id",
            )
        )

        return qs

    @cached_property
    def _page_obj(self):

        if self.fields_by_model:
            history_querysets = [
                self._get_values_qs_for_single_model(m)
                for m in self.fields_by_model.keys()
            ]
        else:
            history_querysets = [
                self._get_values_qs_for_single_model(m) for m in self.models
            ]

        qs_values_union = reduce(self.unionize_querysets, history_querysets)
        sorted_union = qs_values_union.order_by("-business_date")
        paginated_qs = Paginator(sorted_union, self.page_size)
        page = paginated_qs.page(self.page_num)

        return page

    def get_total_entry_count(self):
        return self._page_obj.paginator.count

    def get_total_page_count(self):
        return self._page_obj.paginator.num_pages

    def get_fully_fetched_edit_entries(self):
        """
            our paginated qs only hold id, previous_version_id, eternal_id and model_name in dicts
        """
        slim_versions = self._page_obj.object_list
        models_by_name = {m.__name__: m for m in self.models}

        eternal_ids_to_fetch_by_model = defaultdict(list)
        version_ids_to_fetch_by_model = defaultdict(list)

        for slim_ver in slim_versions:
            eternal_model = models_by_name[slim_ver["model_name"]]
            hist_model = eternal_model._history_class
            eternal_ids_to_fetch_by_model[eternal_model].append(slim_ver["eternal_id"])
            version_ids_to_fetch_by_model[hist_model].append(slim_ver["id"])
            if slim_ver.get("previous_version_id", None):
                version_ids_to_fetch_by_model[hist_model].append(
                    slim_ver["previous_version_id"]
                )

        eternal_records_by_pair_id = {}
        version_records_by_pair_id = {}

        for model, ids in eternal_ids_to_fetch_by_model.items():
            for record in model.objects.filter(id__in=ids):
                eternal_records_by_pair_id[(model, record.id)] = record

        for model, ids in version_ids_to_fetch_by_model.items():
            for record in model.objects.filter(id__in=ids):
                version_records_by_pair_id[(model, record.id)] = record

        resolved_list = []
        for slim_ver in slim_versions:
            eternal_model = models_by_name[slim_ver["model_name"]]
            hist_model = eternal_model._history_class
            resolved = {}

            resolved["eternal"] = eternal_records_by_pair_id[
                (eternal_model, slim_ver["eternal_id"])
            ]

            resolved["version"] = version_records_by_pair_id[(hist_model, slim_ver["id"])]

            if slim_ver.get("previous_version_id", None):
                resolved["previous_version"] = version_records_by_pair_id[
                    (hist_model, slim_ver["previous_version_id"])
                ]
            else:
                resolved["previous_version"] = None

            resolved_list.append(resolved)

        return resolved_list


class SingleRecordConsecutiveVersionsFetcher(ConsecutiveVersionsFetcher):
    def __init__(self, page_size, page_num, model, primary_key=None):
        super().__init__(
            page_num=page_num, page_size=page_size, models=[model],
        )
        self.primary_key = primary_key

    def get_base_version_qs_for_single_model(self, live_model):
        history_model = live_model._history_class
        return history_model.objects.filter(eternal_id=self.primary_key)



import inspect
import logging
import traceback

from graphene.test import Client

from graphql.error import GraphQLError
from graphql.language.base import parse
from graphql.validation import validate

from .graphql_context import GraphQLContext
from .middleware import get_middleware


class GraphQLExecutionErrorSet(Exception):
    def __init__(self, graphql_errors):
        self.graphql_errors = graphql_errors
        # err_str = json.dumps(graphql_errors, indent=4, sort_keys=True)
        super().__init__(graphql_errors)


class RaiseExceptionsMiddleware:
    """
        TODO: this makes every single resolver async, which is probably a performance problem later on when we have more fields. Right now most of our resolvers are async anyway. 

        To fix this, we have to manually create a MiddlewareManager that conditionally wraps async resolvers (by checking asyncio.iscoroutinefunction) with async, and leaves non-async resolvers alone
    """

    logger = logging.getLogger("django.request")

    def on_error(self, error, *args, **kwargs):
        traceback_str = "".join(traceback.format_tb(error.__traceback__))
        self.logger.error(f"{error.__class__.__name__}: {error}")
        self.logger.error(traceback_str)

        raise error

    def resolve(self, next, root, info, **kwargs):
        return next(root, info, **kwargs).catch(self.on_error)


class InternalQueryExecutorBase:
    """
        - must define class variable "schema"

        if you want to execute multiple graphQL queries against the same data-loaders
        you must re-use instances of this class
    """

    schema = None

    def __init__(self):
        self.client = Client(self.schema)
        self.dataloaders = {}

    def execute_query(
        self, query: str, root: any = None, context: dict = None, variables: dict = None
    ):
        if context is None:
            context = GraphQLContext(self.dataloaders, requires_serializable=False)

        middleware = [RaiseExceptionsMiddleware(), *get_middleware()]

        resp = self.client.execute(query, root, context, variables, middleware=middleware)

        if "errors" in resp:
            err = GraphQLExecutionErrorSet(resp["errors"])
            raise err
        return resp["data"]

    def build_query(self, query):
        validation_errors = validate(self.schema, parse(query))
        if validation_errors:
            raise GraphQLError(validation_errors)

        def execute(*, context=None, **variables):
            return self.execute_query(query, context, variables=variables)

        return execute



from django_sample.models import Author, Book, Tag
from zeus.changelog.simple_changelog import create_simple_changelog


def test_simple_changelog():
    author1 = Author.objects.create(first_name="john", last_name="smith")
    author2 = Author.objects.create(first_name="jane", last_name="smith")
    book1 = Book.objects.create(author=author1, title="john's diary")

    book1_v1 = book1.versions.last()

    # refresh a new copy of this record to avoid history-row re-use
    book1.reset_version_attrs()
    book1.title = "jane's diary"
    book1.author = author2
    book1.save()

    book1_v2 = book1.versions.last()

    changelog = create_simple_changelog(models=[Author, Book], page_size=2)

    data = changelog.get_page(1)
    edit_entries = data["changelog_entries"]
    assert len(edit_entries) == 2

    assert edit_entries[0]["version"]["instance"] == book1_v2
    assert len(edit_entries[0]["diffs"]) == 2
    diffs_by_fields = {diff["field_name"]: diff for diff in edit_entries[0]["diffs"]}

    # field-diffs might come in any order
    assert set(diffs_by_fields.keys()) == {
        "title",
        "author",
    }
    # they have field objects as well
    assert diffs_by_fields["title"]["field"] == Book._meta.get_field("title")

    assert edit_entries[1]["version"]["instance"] == book1_v1

    # this field returns a tm() result
    assert edit_entries[1]["diffs"][0]["action"] == "Created"

    assert edit_entries[0]["eternal"] == edit_entries[1]["eternal"] == book1
    assert edit_entries[0]["model_name"] == "book"

    assert data["has_next_page"] == True



from django.db.models import ForeignKey, ManyToManyField

import graphene
from graphene import DateTime, List, String

from zeus.changelog.util import get_diffable_fields_for_model
from zeus.graphql.types.version import Version
from zeus.graphql.utils import NonSerializable, non_serializable_field

from .changelog_entry_field_entry import ChangelogEntryFieldEntry
from .diff import CreateDiff, Diff, get_field_diff_for_version_pair


class ChangelogEntry(graphene.ObjectType):
    diffs = List(Diff,)
    field_entries = List(ChangelogEntryFieldEntry)
    model_name = graphene.String()
    live_name = graphene.String()
    edit_date = graphene.DateTime()

    eternal = NonSerializable()
    version = graphene.Field(Version)
    previous_version = graphene.Field(Version)

    is_annotated = graphene.Boolean()

    @staticmethod
    def resolve_version(parent, _info):
        return {"instance": parent["version"]}

    @staticmethod
    def resolve_previous_version(parent, _info):
        if not parent["previous_version"]:
            return None
        return {"instance": parent["previous_version"]}

    @staticmethod
    @non_serializable_field
    def resolve_eternal(parent, _info):
        return parent["eternal"]

    @staticmethod
    def resolve_edit_date(parent, _info):
        return parent["version"].business_date

    @staticmethod
    def resolve_model_name(parent, _info):
        return parent["eternal"].__class__._meta.verbose_name

    @staticmethod
    def resolve_live_name(parent, _info):
        live_obj = parent["eternal"]
        if hasattr(live_obj, "name"):
            return live_obj.name

        return live_obj.__str__()

    @staticmethod
    def resolve_is_annotated(parent, _info):
        version = parent["version"]

        is_cosmestic_change = version.cosmetic_change
        has_either_lang_filled_out = bool(
            version.reason_for_substantive_change_en
            or version.reason_for_substantive_change_fr
        )
        is_creation = parent["previous_version"] is None

        return is_cosmestic_change or has_either_lang_filled_out or is_creation

    @staticmethod
    def resolve_diffs(parent, info):
        specified_fields = parent.get("fields", None)
        this_version = parent["version"]
        prev_version = parent["previous_version"]

        if prev_version is None:
            # filter out creations if fields are specified
            if specified_fields:
                return []
            eternal = parent["eternal"]
            return [CreateDiff()]

        fields_to_diff = get_diffable_fields_for_model(this_version.live_model)
        if specified_fields:
            fields_to_diff = [f for f in fields_to_diff if f.name in specified_fields]

        field_diff_objs = []
        for f in fields_to_diff:
            field_diff_obj = get_field_diff_for_version_pair(
                this_version, prev_version, f, info.context.dataloaders
            )
            if field_diff_obj is not None:
                field_diff_objs.append(field_diff_obj)

        return field_diff_objs

    @staticmethod
    def resolve_field_entries(parent, _info):
        field_objs = get_diffable_fields_for_model(parent["version"].live_model)

        return [
            {
                "version": parent["version"],
                "previous_version": parent["previous_version"],
                "field": field,
            }
            for field in field_objs
        ]



import graphene

from zeus.changelog.graphql.util import (
    convert_enum_list_to_dict,
    create_model_enum_type,
    create_model_field_enum_type,
    create_standard_changelog_graphql_mixin,
)
from zeus.graphql.internal_query_executor_base import InternalQueryExecutorBase

base_query = """
    query ChangelogQuery(
        $page_num :Int!
        $user_ids: [Int]
        $models:[ChangelogModels]
        $fields: [ChangelogModelFields]
        $exclude_create: Boolean!
        $only_creates: Boolean!
        $start_date: DateTime
        $end_date: DateTime
    ) {
        changelog(
            page_num:$page_num,
            user_ids: $user_ids,
            models: $models,
            fields: $fields,
            exclude_create: $exclude_create,
            only_creates: $only_creates,
            start_date: $start_date,
            end_date: $end_date,
        ){
            num_pages
            has_next_page
            changelog_entries {
                model_name
                version {
                    instance
                    edited_by
                }
                eternal
                live_name
                diffs {
                    field
                    field_name
                    action
                    diffed_before
                    diffed_after
                    diffed_combined
                }
            }
        }
    }
"""


def create_simple_changelog(models, page_size=None):
    RootQuery = create_standard_changelog_graphql_mixin(
        diffable_models=models, page_size=page_size
    )
    changelog_schema = graphene.Schema(query=RootQuery, auto_camelcase=False)

    class QueryExecutor(InternalQueryExecutorBase):
        schema = changelog_schema

    query_executor = QueryExecutor()

    return ChangelogContainer(query_executor)


class ChangelogContainer:
    def __init__(self, query_executor):
        self.query_executor = query_executor

    def get_page(
        self,
        page_num,
        models=None,
        user_ids=None,
        fields=None,
        exclude_create=False,
        only_creates=False,
        start_date=None,
        end_date=None,
    ):

        page = self.query_executor.execute_query(
            base_query,
            variables={
                "page_num": page_num,
                "models": models,
                "user_ids": user_ids,
                "fields": fields,
                "exclude_create": exclude_create,
                "only_creates": only_creates,
                "start_date": start_date,
                "end_date": end_date,
            },
        )
        return page["changelog"]



import pytest


@pytest.fixture(autouse=True)
def enable_db_access_for_all_tests(db):
    """
    without this, tests (including old-style) have to explicitly delcare db as a dependency
    this fixture is extremely magic, but it's what I found here 
    https://pytest-django.readthedocs.io/en/latest/faq.html#how-can-i-give-database-access-to-all-my-tests-without-the-django-db-marker
  """
    pass



from django.urls import reverse
from django.views.generic import TemplateView


class AbstractChangelogView(TemplateView):
    """
        query must contain 
            changelog(page_num,...)
                num_pages
                has_next_page
                changelog_entries {
            
                }
            }

    """

    template_name = None
    graphql_query = None
    query_executor_class = None

    @classmethod
    def as_view(cls, *args, **kwargs):
        assert cls.template_name, "must define template_name"
        assert cls.graphql_query, "must define graphql_query"
        assert cls.query_executor_class, "must define query_executor_class"
        return super().as_view(*args, **kwargs)

    def get_graphql_variables(self):
        page_num = self.kwargs.get("page_num", 1)
        return {
            "page_num": page_num,
        }

    def get_changelog_data(self):
        page_num = self.kwargs.get("page_num", 1)
        variables = self.get_graphql_variables()
        query_executor = self.query_executor_class()
        data = query_executor.execute_query(self.graphql_query, variables=variables)

        edit_entries = data["changelog"]["changelog_entries"]
        has_next_page = data["changelog"]["has_next_page"]
        num_pages = data["changelog"]["num_pages"]

        entries_without_diffs = [entry for entry in edit_entries if not entry["diffs"]]

        prev_page = None
        next_page = None

        if page_num > 1:
            prev_page = page_num - 1

        if has_next_page:
            next_page = page_num + 1

        return {
            "entries_without_diffs": entries_without_diffs,
            "edit_entries": edit_entries,
            "prev_page_num": prev_page,
            "next_page_num": next_page,
            "num_pages": num_pages,
            "page_num": page_num,
        }

    def get_context_data(self, *args, **kwargs):
        changelog_context_data = self.get_changelog_data()
        return {
            **super().get_context_data(*args, **kwargs),
            **changelog_context_data,
        }



from django.db import connection


class QueryCollector:
    """
        usage:

        from django.db import connection
        qc = QueryCollector()
        with connection.execute_wrapper(qc):
            do_queries()
        # Now we can print the log.
        print(qc.queries)
    """

    def __init__(self):
        self.queries = []

    def __call__(self, execute, sql, params, many, context):
        current_query = {"sql": sql, "params": params, "many": many}
        try:
            result = execute(sql, params, many, context)
        except Exception as e:
            current_query["status"] = "error"
            current_query["exception"] = e
            raise
        else:
            current_query["status"] = "ok"
            return result
        finally:
            self.queries.append(current_query)


class QueryCounter:
    """
        usage: 
            with QueryCounter() as qc:
                # arbitrary code that execs sql 
            
            assert( qc.count < 10 )
    """

    def __init__(self):
        self.query_collector = QueryCollector()
        self.connection_ctx_manager = connection.execute_wrapper(self.query_collector)

    def __enter__(self):
        self.connection_ctx_manager.__enter__()

    def __exit__(self, *excp):
        self.connection_ctx_manager.__exit__(*excp)
        self.count = len(self.query_collector.queries)


class MaxQueryCountExceededError(Exception):
    pass


class assert_max_queries:
    """
        usage:

        with assert_max_queries(n):
            ... code_that_should_make_up_to_n_queries
    """

    def __init__(self, max_queries: int):
        self.qc = QueryCounter()
        self.max_queries = max_queries

    def __enter__(self):
        self.qc.__enter__()

    def __exit__(self, *excp):
        self.qc.__exit__(*excp)
        if self.qc.count > self.max_queries:
            raise MaxQueryCountExceededError(
                f"Expected a maximum of {self.max_queries} but got {self.qc.count} queries"
            )



from django.utils.safestring import mark_safe

import bleach
import mistune

ALLOWED_TAGS = ["a", "b", "em", "i", "li", "ol", "p", "strong", "ul", "br", "div"]


def sanitize_html(html_str, allow_weird_characters=False):
    with_terrible_replacements = bleach.clean(html_str, tags=ALLOWED_TAGS)
    # see https://github.com/mozilla/bleach/issues/192
    if allow_weird_characters:
        return with_terrible_replacements.replace("&amp;", "&")
    else:
        return with_terrible_replacements


def is_md_valid(md_str):
    markdown_from_src = mistune.Markdown()
    unsanitized = markdown_from_src(str(md_str))
    sanitized = sanitize_html(unsanitized)
    return unsanitized == sanitized


def markdown(md_str):
    """
    mark_safe being run through bleach
  """
    markdown = mistune.Markdown()
    return mark_safe(sanitize_html(markdown(str(md_str))))



import functools
import inspect
from inspect import isawaitable
from django.db.models import Model
from asgiref.sync import sync_to_async


import graphene

from promise import is_thenable


class NonSerializable(graphene.Scalar):
    @staticmethod
    def serialize(obj):
        return obj


def _stringify_internal_python_value(val):
    if isinstance(val, Model):
        # if the model's __str__ methods issues queries
        # it can be misleading when investigating N+1 queries with graphiql
        return f"<{val.__class__.__name__} {val.pk}>"

    return object.__str__(val)


def non_serializable_field(func):
    """
        decorate a resolver
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)

        info = args[1]
        if len(args) == 3:
            # if we're dealing with a class or instance method, info will be third
            info = args[2]

        if info.context.requires_serializable:
            if is_thenable(ret):
                return ret.then(lambda val: _stringify_internal_python_value(val))

            return _stringify_internal_python_value(ret)

        return ret

    return wrapper


class HasNonSerializableRecordMixin(graphene.ObjectType):
    record = graphene.Field(NonSerializable)

    def resolve_record(self, _info):
        return self


def genfunc_to_prom(func):
    """
        turns a function returning a generator into a function returning a promise
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        generator = func(*args, **kwargs)
        if not inspect.isgenerator(generator):
            return generator

        return promise_from_generator(generator)

    return wrapper


def promise_from_generator(generator):
    """
        assumes the generator is not yet started
    """
    try:
        first_val = next(generator)
    except StopIteration as e:
        next_val = e.value
        return next_val

    return _ongoing_gen_to_prom(generator, first_val)


def _ongoing_gen_to_prom(generator, current_val=None):
    """
        recursive helper
    """
    if inspect.isgenerator(current_val):
        current_val = promise_from_generator(current_val)
    if not is_thenable(current_val):
        # this must be the final return
        return current_val

    def resolve_next(resolved_current):
        try:
            next_val = generator.send(resolved_current)
        except StopIteration as e:
            next_val = e.value
            return next_val

        return _ongoing_gen_to_prom(generator, next_val)

    return current_val.then(resolve_next)


def genfunc_to_coroutine(func):
    """
        turns a function returning a generator into a coroutine
    """

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        # result = func(*args, **kwargs)
        wrapped_func = sync_to_async(func)
        result = await wrapped_func(*args, **kwargs)
        if inspect.isgenerator(result):
            return awaitable_from_generator(result)

        return result

    return wrapper


def awaitable_from_generator(generator):
    """
        assumes the generator is not yet started
    """
    try:
        first_val = next(generator)
    except StopIteration as e:
        next_val = e.value
        return next_val

    return _ongoing_gen_to_awaitable(generator, first_val)


async def _ongoing_gen_to_awaitable(generator, current_val=None):
    """
        recursive helper
    """
    if inspect.isgenerator(current_val):
        current_val = awaitable_from_generator(current_val)
    if not isawaitable(current_val):
        # this must be the final return
        return current_val

    def resolve_next(resolved_current):
        try:
            next_val = generator.send(resolved_current)
        except StopIteration as e:
            next_val = e.value
            return next_val

        return _ongoing_gen_to_awaitable(generator, next_val)

    current_val = await current_val
    return resolve_next(current_val)



import sys

from django.conf import settings
from django.utils.functional import lazy
from django.utils.safestring import mark_safe
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _

from zeus.markdown import markdown, sanitize_html

from .text_entry_manager import TextEntryManager


class TextMakerError(Exception):
    pass


class TextMakerCreator:
    def __init__(self, global_keys, text_files):
        self.validate_text_files(text_files)

        self.entry_manager = TextEntryManager(text_files)
        self.entry_manager.load()
        self.global_keys = global_keys

    @staticmethod
    def validate_text_files(text_files):
        for path in text_files:
            if path.startswith("."):
                raise TextMakerError(f"don't start paths with './' offender: {path}")
            if not path.endswith(".text.yaml"):
                raise TextMakerError("files must have the '.text.yaml' extension")

    def get_tm_func(self):
        return lazy(self._tm, str)

    def _tm(self, key, allow_md=True, extra_keys={}):
        lang = (
            get_language() or "en"
        )  # when called from a non-localized URL (e.g. admin), get_language() is None
        global_keys = self.global_keys[lang]

        if key in global_keys:
            return global_keys[key]

        elif self.entry_manager.has_entry(key):
            entry = self.entry_manager.get_entry(key)

        else:
            raise TextMakerError(f"text key {key} doesn't exist")

        if allow_md is False and entry.get("md", False):
            raise TextMakerError("requested key contains markdown")

        # having unsafe: True on a yaml entry allows weird characters (such as &) to be unescaped
        # note that HTML sanitization will still occur in this case
        # this should be used anytime we interpolate user-input into a yaml template string
        should_mark_safe = True
        if entry.get("unsafe", None):
            should_mark_safe = False

        try:
            if lang in entry:
                text = entry[lang] or f"FIXME: {key}"
            else:
                text = f"FIXME: {entry['en']}"

            if entry.get("md"):
                text = markdown(text)

            template_args = {**global_keys, **extra_keys}

            text = text % template_args

            sanitized = sanitize_html(text, allow_weird_characters=not (should_mark_safe))

            if should_mark_safe:
                return mark_safe(sanitized)
            else:
                return sanitized
        except Exception as e:
            raise TextMakerError(
                f"text_key {key} with arguments: {extra_keys} had the following error \n {e}"
            )


class WatchingTextMakerCreator(TextMakerCreator):
    def __init__(self, global_keys, text_files):
        super().__init__(global_keys, text_files)

        if settings.DEBUG and sys.argv[1] in ("runserver", "rs"):
            self.start_watch()

    def start_watch(self):
        # prevent
        # pylint: disable="import-outside-toplevel relative-beyond-top-level"
        from .text_watcher import text_watcher

        text_watcher.add_entry_watcher(self.entry_manager)
        text_watcher.start_watching()



import asyncio
import logging
import traceback

from django.conf import settings
from django.core.exceptions import SuspiciousOperation

from graphene_django.views import GraphQLView

from .graphql_context import GraphQLContext
from .middleware import get_middleware


class GraphiQLViewBase(GraphQLView):
    """
        - must provide class attr schema
    """

    logger = logging.getLogger("django.request")
    schema = None

    def __init__(self, *args, executor=None, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def as_view(cls, *args, **kwargs):
        new_kwargs = {
            "schema": cls.schema,
            "graphiql": True,
            "middleware": get_middleware(),
            **kwargs,
        }
        return super().as_view(*args, **new_kwargs)

    def dispatch(self, *args, **kwargs):
        if not settings.DEBUG:
            raise SuspiciousOperation("this is a dev-only feature")

        return super().dispatch(*args, **kwargs)

    def get_context(self, request):
        dataloaders = {}
        return GraphQLContext(dataloaders, requires_serializable=True)

    def execute_graphql_request(self, *args, **kwargs):
        result = super().execute_graphql_request(*args, **kwargs)
        if result.errors:
            self._log_exceptions(result.errors)
        return result

    def _log_exceptions(self, errors):
        for error in errors:
            error_to_log = error
            if hasattr(error, "original_error"):
                error_to_log = error.original_error
            traceback_str = "".join(traceback.format_tb(error_to_log.__traceback__))
            self.logger.error(f"{error_to_log.__class__.__name__}: {error_to_log}")
            self.logger.error(traceback_str)



from django.contrib.auth.models import User
from django.urls import reverse

import pytest

from django_sample.models import Author, Book


@pytest.fixture
def data_url():
    author = Author.objects.create(first_name="john")
    a_book = Book.objects.create(author=author, title="name1")
    url = reverse("edit-book", args=[a_book.pk])
    return url


def test_unauthenticated_request_redirects_to_login(client, data_url):
    resp = client.get(data_url, follow=True)
    assert resp.resolver_match.url_name == "login"


def test_authenticated_request_is_ok(client, data_url):
    user = User.objects.create(username="abc")
    client.force_login(user)
    resp = client.get(data_url)
    assert resp.status_code == 200



class disconnect_signal:
    """ Temporarily disconnect a model from a signal """

    def __init__(self, signal, receiver, sender=None, dispatch_uid=None):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender
        self.dispatch_uid = dispatch_uid

    def __enter__(self):
        self.signal.disconnect(
            receiver=self.receiver, sender=self.sender, dispatch_uid=self.dispatch_uid,
        )

    def __exit__(self, type, value, traceback):
        self.signal.connect(
            receiver=self.receiver,
            sender=self.sender,
            dispatch_uid=self.dispatch_uid,
            weak=False,
        )



import pytest
from promise import get_default_scheduler, set_default_scheduler
from promise.schedulers.asyncio import AsyncioScheduler


@pytest.fixture
def dataloader_test():
    """
        allows promises not to be executed immediately during test runs
    """
    # For some reason, dataloaders .load() method instantly call batch_load when testing, making it impossible to test batching and counting queries.
    # Changing the scheduler during dataloader tests seems to fix the issue without breaking trigerring any async exception
    default_scheduler = get_default_scheduler()
    set_default_scheduler(AsyncioScheduler())
    yield
    set_default_scheduler(default_scheduler)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="titan-zeus",
    version="0b0",
    author="AlexCLeduc",
    # author_email="author@example.com",
    # description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TBS-EACPD/zeus",
    packages=[
        # find_packages() also includes extraneous stuff, like testing and django_sample
        package
        for package in setuptools.find_packages()
        if package.startswith("zeus")
    ],
    package_data={"": ["**/*.text.yaml"]},
    include_package_data=True,
    install_requires=[],
    tests_require=["django"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import graphene
from graphene import Boolean, Int, List

from .changelog_entry import ChangelogEntry


class ChangelogPage(graphene.ObjectType):
    has_next_page = Boolean()
    total_version_count = Int()
    num_pages = Int()
    changelog_entries = List(ChangelogEntry)

    @staticmethod
    def resolve_changelog_entries(parent, _info):
        fetcher = parent["fetcher"]
        fields_by_model = fetcher.fields_by_model or {}
        return [
            {
                **data_record,
                "fields": fields_by_model.get(data_record["eternal"].__class__, None),
            }
            for data_record in fetcher.get_fully_fetched_edit_entries()
        ]

    @staticmethod
    def resolve_has_next_page(parent, _info):
        fetcher = parent["fetcher"]
        return fetcher.get_total_page_count() > fetcher.page_num

    @staticmethod
    def resolve_total_version_count(parent, _info):
        fetcher = parent["fetcher"]
        return fetcher.get_total_entry_count() > fetcher.page_num

    @staticmethod
    def resolve_num_pages(parent, _info):
        fetcher = parent["fetcher"]
        return fetcher.get_total_page_count()



from collections import defaultdict

import graphene
from graphene import Boolean, DateTime, Int, List

from django_sample.models import Author, Book
from zeus.changelog.arbitrary_version_pair_fetcher import ArbitraryVersionPairFetcher
from zeus.changelog.consecutive_versions_fetcher import (
    ConsecutiveVersionsFetcher,
    SingleRecordConsecutiveVersionsFetcher,
)
from zeus.changelog.graphql.types import (
    ChangelogEntry,
    ChangelogPage,
    VersionComparisonEntry,
)
from zeus.changelog.graphql.util import (
    convert_enum_list_to_dict,
    create_model_enum_type,
    create_model_field_enum_type,
    create_standard_changelog_graphql_mixin,
)
from zeus.graphql.internal_query_executor_base import InternalQueryExecutorBase

diffable_models = [
    Author,
    Book,
]
RootQuery = create_standard_changelog_graphql_mixin(diffable_models=diffable_models)

schema = graphene.Schema(query=RootQuery, auto_camelcase=False)


class QueryExecutor(InternalQueryExecutorBase):
    schema = schema



from functools import reduce

from django.db.models import (
    BooleanField,
    CharField,
    F,
    IntegerField,
    OuterRef,
    Q,
    Subquery,
    Value,
)
from django.utils.functional import cached_property

from .util import get_diffable_fields_for_model


class ArbitraryVersionPairFetcher:
    """
        given a list of queryset pairs (one pair per model)
        outputs a list of {right_id, left_id, eternal_id, model}
        automatically filters out record where right_id == left_id
    """

    def __init__(self, qs_pairs):
        self.qs_pairs = qs_pairs

    @cached_property
    def models_by_name(self):
        models_by_name = {
            qs_pair[0].model.live_model.__name__: qs_pair[0].model
            for qs_pair in self.qs_pairs
        }
        return models_by_name

    def _get_union_for_models(self, querysets):
        model_querysets = [
            qs.annotate(
                model_name=Value(qs.model.live_model.__name__, CharField())
            ).values("model_name", "id", "eternal_id")
            for qs in querysets
        ]

        unionize = lambda qs1, qs2: qs1.union(qs2)
        unioned_qs = reduce(unionize, model_querysets)
        return unioned_qs

    def _index_qs(self, qs):
        indexed = {}
        for record in qs:
            model = self.models_by_name[record["model_name"]]
            eternal_id = record["eternal_id"]
            indexed[(model, eternal_id)] = record["id"]

        return indexed

    def get_version_pairs(self):
        left_querysets = [p[0] for p in self.qs_pairs]
        right_querysets = [p[1] for p in self.qs_pairs]

        left_union = self._get_union_for_models(left_querysets)
        right_union = self._get_union_for_models(right_querysets)

        left_minus_right = self._index_qs(left_union.difference(right_union))
        right_minus_left = self._index_qs(right_union.difference(left_union))

        final_values = []
        for model, eternal_id in {*left_minus_right.keys(), *right_minus_left.keys()}:
            right_id = right_minus_left.get((model, eternal_id), None)
            left_id = left_minus_right.get((model, eternal_id), None)
            final_values.append(
                {
                    "right_id": right_id,
                    "left_id": left_id,
                    "model": model,
                    "eternal_id": eternal_id,
                }
            )

        return final_values
